@@ -27,7 +27,6 @@ import gst
 
 from pitivi.log.loggable import Loggable
 from pitivi.elements.singledecodebin import SingleDecodeBin
-from pitivi.elements.smartscale import SmartVideoScale
 from pitivi.signalinterface import Signallable
 from pitivi.stream import match_stream_groups_map, AudioStream, VideoStream
 from pitivi.utils import formatPercent
@@ -323,35 +322,39 @@ class SourceFactory(ObjectFactory):
         bin.remove_pad(ghost_pad)
 
     def _releaseBin(self, bin):
-        try:
-            # bin is a bin returned from makeDefaultBin
-            bin.decodebin.disconnect_by_func(self._binNewDecodedPadCb)
-            bin.decodebin.disconnect_by_func(self._binRemovedDecodedPadCb)
-        except TypeError:
-            # bin is a stream bin
-            bin.decodebin.disconnect_by_func(self._singlePadAddedCb)
-            bin.decodebin.disconnect_by_func(self._singlePadRemovedCb)
+        if hasattr(bin, "decodebin"):
+            try:
+                # bin is a bin returned from makeDefaultBin
+                bin.decodebin.disconnect_by_func(self._binNewDecodedPadCb)
+                bin.decodebin.disconnect_by_func(self._binRemovedDecodedPadCb)
+            except TypeError:
+                # bin is a stream bin
+                bin.decodebin.disconnect_by_func(self._singlePadAddedCb)
+                bin.decodebin.disconnect_by_func(self._singlePadRemovedCb)
+            del bin.decodebin
 
-        del bin.decodebin
         if hasattr(bin, "child"):
             bin.child.set_state(gst.STATE_NULL)
             del bin.child
 
         if hasattr(bin, "volume"):
             # only audio bins have a volume element
-            for elt in [bin.aconv, bin.ares, bin.volume]:
+            for elt in [bin.aconv, bin.ares, bin.arate, bin.volume]:
                 elt.set_state(gst.STATE_NULL)
                 bin.remove(elt)
             del bin.volume
             del bin.aconv
             del bin.ares
+            del bin.arate
         elif hasattr(bin, "alpha"):
-            for elt in [bin.csp, bin.queue, bin.alpha]:
+            for elt in [bin.csp, bin.queue, bin.alpha, bin.capsfilter, bin.scale]:
                 elt.set_state(gst.STATE_NULL)
                 bin.remove(elt)
             del bin.queue
             del bin.csp
             del bin.alpha
+            del bin.capsfilter
+            del bin.scale
 
         if hasattr(bin, "ghostpad"):
             # singledecodebin found something on this pad
@@ -359,13 +362,17 @@ class SourceFactory(ObjectFactory):
             bin.remove_pad(bin.ghostpad)
             del bin.ghostpad
 
-    def _makeStreamBin(self, output_stream, child_bin=None):
-        self.debug("output_stream:%r", output_stream)
+    def _makeStreamBinReal(self, output_stream):
         b = gst.Bin()
         b.decodebin = self.singleDecodeBinClass(uri=self.uri, caps=output_stream.caps,
                                            stream=output_stream)
         b.decodebin.connect("pad-added", self._singlePadAddedCb, b)
         b.decodebin.connect("pad-removed", self._singlePadRemovedCb, b)
+        return b
+
+    def _makeStreamBin(self, output_stream, child_bin=None):
+        self.debug("output_stream:%r", output_stream)
+        b = self._makeStreamBinReal(output_stream)
         if child_bin:
             b.child = child_bin
             b.add(child_bin)
@@ -375,16 +382,20 @@ class SourceFactory(ObjectFactory):
             # add a volume element
             b.aconv = gst.element_factory_make("audioconvert", "internal-aconv")
             b.ares = gst.element_factory_make("audioresample", "internal-audioresample")
+            # Fix audio jitter of up to 40ms
+            b.arate = gst.element_factory_make("audiorate", "internal-audiorate")
+            b.arate.props.tolerance = 40 * gst.MSECOND
             b.volume = gst.element_factory_make("volume", "internal-volume")
-            b.add(b.volume, b.ares, b.aconv)
+            b.add(b.volume, b.ares, b.aconv, b.arate)
             if child_bin:
-                gst.element_link_many(b.aconv, b.ares, b.child, b.volume)
+                gst.element_link_many(b.aconv, b.ares, b.arate, b.child, b.volume)
                 b.child.sync_state_with_parent()
             else:
-                gst.element_link_many(b.aconv, b.ares, b.volume)
-                
+                gst.element_link_many(b.aconv, b.ares, b.arate, b.volume)
+
             b.aconv.sync_state_with_parent()
             b.ares.sync_state_with_parent()
+            b.arate.sync_state_with_parent()
             b.volume.sync_state_with_parent()
         elif isinstance(output_stream, VideoStream):
             self.debug("Adding alpha element")
@@ -410,23 +421,31 @@ class SourceFactory(ObjectFactory):
                 b.csp = gst.element_factory_make("identity")
 
             b.alpha = gst.element_factory_make("alpha", "internal-alpha")
-            b.scale = SmartVideoScale()
-            b.scale.set_caps(self._filtercaps)
-            b.scale._computeAndSetValues()
+            b.alpha.props.prefer_passthrough = True
+            b.scale = gst.element_factory_make("videoscale")
+            try:
+                b.scale.props.add_borders = True
+            except AttributeError:
+                self.warning("User has old version of videoscale. "
+                        "add-border not enabled")
+            b.capsfilter = gst.element_factory_make("capsfilter")
+            self.setFilterCaps(self._filtercaps, b)
 
-            b.add(b.queue, b.scale, b.csp, b.alpha)
+            b.add(b.queue, b.scale, b.csp, b.alpha, b.capsfilter)
             gst.element_link_many(b.queue, b.csp, b.scale)
             if child_bin:
-                gst.element_link_many(b.scale, b.child, b.alpha)
+                gst.element_link_many(b.scale, b.child, b.alpha, b.capsfilter)
                 b.child.sync_state_with_parent()
             else:
-                gst.element_link_many(b.scale, b.alpha)
+                gst.element_link_many(b.scale, b.alpha, b.capsfilter)
+            b.capsfilter.sync_state_with_parent()
             b.scale.sync_state_with_parent()
             b.queue.sync_state_with_parent()
             b.csp.sync_state_with_parent()
             b.alpha.sync_state_with_parent()
 
-        b.add(b.decodebin)
+        if hasattr(b, "decodebin"):
+            b.add(b.decodebin)
         return b
 
     def _singlePadAddedCb(self, dbin, pad, topbin):
@@ -438,17 +457,17 @@ class SourceFactory(ObjectFactory):
             # since those elements are still unlinked downstream at this point,
             # so state change order doesn't happen in the usual
             # downstream-to-upstream way.
-            for element in [topbin.aconv, topbin.ares, topbin.volume]:
+            for element in [topbin.aconv, topbin.ares, topbin.arate, topbin.volume]:
                 element.sync_state_with_parent()
 
             pad.link(topbin.aconv.get_pad("sink"))
             topbin.ghostpad = gst.GhostPad("src", topbin.volume.get_pad("src"))
         elif hasattr(topbin, "alpha"):
-            for element in [topbin.queue, topbin.scale, topbin.csp, topbin.alpha]:
+            for element in [topbin.queue, topbin.scale, topbin.csp, topbin.alpha, topbin.capsfilter]:
                 element.sync_state_with_parent()
 
             pad.link(topbin.queue.get_pad("sink"))
-            topbin.ghostpad = gst.GhostPad("src", topbin.alpha.get_pad("src"))
+            topbin.ghostpad = gst.GhostPad("src", topbin.capsfilter.get_pad("src"))
         else:
             topbin.ghostpad = gst.GhostPad("src", pad)
 
@@ -476,12 +495,22 @@ class SourceFactory(ObjectFactory):
     def addInputStream(self, stream):
         raise AssertionError("source factories can't have input streams")
 
-    def setFilterCaps(self, caps):
-        self._filtercaps = caps
-        for b in self.bins:
-            if hasattr(b, "scale"):
-                b.scale.set_caps(caps)
-                b.scale._computeAndSetValues()
+    def setFilterCaps(self, caps, b=None):
+        caps_copy = gst.Caps(caps)
+        for structure in caps_copy:
+            # remove framerate as we don't adjust framerate here
+            if structure.has_key("framerate"):
+                del structure["framerate"]
+            # remove format as we will have converted to AYUV/ARGB
+            if structure.has_key("format"):
+                del structure["format"]
+        if b is None:
+            for bin in self.bins:
+                if hasattr(bin, "capsfilter"):
+                    bin.capsfilter.props.caps = caps_copy
+        else:
+            b.capsfilter.props.caps = caps_copy
+        self._filtercaps = caps_copy
 
 class SinkFactory(ObjectFactory):
     """
