@@ -132,8 +132,8 @@ class Interpolator(Signallable, Loggable):
 
     __signals__ = {
         'keyframe-added' : ['keyframe'],
-        'keyframe-removed' : ['keyframe'],
-        'keyframe-moved' : ['keyframe'],
+        'keyframe-removed' : ['keyframe', 'old_value'],
+        'keyframe-moved' : ['keyframe', 'old_value'],
     }
 
     def __init__(self, trackobject, element, prop, minimum=None, maximum=None,
@@ -209,6 +209,7 @@ class Interpolator(Signallable, Loggable):
                    gst.TIME_ARGS(keyframe.time), keyframe.value, keyframe.mode)
 
         self._keyframes.append(keyframe)
+        self._keyframes.sort()
 
         self._controller.set(self._property.name, keyframe.time, keyframe.value)
 
@@ -217,10 +218,11 @@ class Interpolator(Signallable, Loggable):
         return keyframe
 
     def removeKeyframe(self, keyframe):
+        old_value = self._controller.get(self._property.name, keyframe.time)
         self._controller.unset(self._property.name, keyframe.time)
         if keyframe is not self.start and keyframe is not self.end:
             self._keyframes.remove(keyframe)
-            self.emit("keyframe-removed", keyframe)
+            self.emit("keyframe-removed", keyframe, old_value)
 
     def setKeyframeMode(self, kf, mode):
         # FIXME: currently InterpolationSourceControllers only support a
@@ -246,23 +248,21 @@ class Interpolator(Signallable, Loggable):
         self.debug("kf.time:%s, ptime:%s, value:%r",
                    gst.TIME_ARGS(kf.time),
                    gst.TIME_ARGS(ptime), value)
+        old_value = self._controller.get(self._property.name, ptime)
         self._controller.set(self._property.name, ptime, value)
         if kf.time != ptime:
             self._controller.unset(self._property.name, kf.time)
-        self.emit("keyframe-moved", kf)
+        self.emit("keyframe-moved", kf, old_value)
 
     def getKeyframes(self):
-        # TODO: This could be more efficient. We are re-sorting all the keyframes
-        # every time we iterate over them. One thought would be to keep the
-        # list sorted in-place as keyframes are added and moved.
         yield self.start
-        for kf in sorted(self._keyframes):
+        for kf in self._keyframes:
             yield kf
         yield self.end
 
     def getInteriorKeyframes(self):
         """Same as above but does not include start, or end points"""
-        for kf in sorted(self._keyframes):
+        for kf in self._keyframes:
             yield kf
 
     def getVisibleKeyframes(self):
@@ -301,6 +301,7 @@ class TrackObject(Signallable, Loggable):
         'media-duration-changed': ['media-duration'],
         'priority-changed': ['priority'],
         'selected-changed' : ['state'],
+        'stagger-changed' : ['stagger'],
     }
 
     def __init__(self, factory, stream, start=0,
@@ -310,10 +311,14 @@ class TrackObject(Signallable, Loggable):
         self.debug("factory:%r", factory)
         self.factory = factory
         self.stream = stream
+        self.stream_type = type(stream)
         self.track = None
         self.timeline_object = None
         self.interpolators = {}
-        self._rebuild_interpolators = True
+        self._rebuild_interpolators = False
+        self._public_priority = priority
+        self._position = 0
+        self._stagger = 0
         self.gnl_object = obj = self._makeGnlObject()
         self.keyframes = []
 
@@ -337,6 +342,7 @@ class TrackObject(Signallable, Loggable):
         obj.props.priority = priority
 
         self._connectToSignals(obj)
+        self._updatePriority(self._public_priority)
 
     def getInterpolator(self, property_name):
         self._maybeBuildInterpolators()
@@ -351,12 +357,12 @@ class TrackObject(Signallable, Loggable):
         return self.interpolators
 
     def _maybeBuildInterpolators(self):
-        if not list(self.gnl_object.elements()):
-            raise TrackError("makeBin hasn't been called yet")
-
         if not self._rebuild_interpolators:
             return
 
+        if not list(self.gnl_object.elements()):
+            raise TrackError("makeBin hasn't been called yet")
+
         self._rebuild_interpolators = False
 
         factory_properties = self.factory.getInterpolatedProperties(self.stream)
@@ -379,14 +385,17 @@ class TrackObject(Signallable, Loggable):
                 interpolator = Interpolator(self, gst_object,
                     gst_object_property, lower, upper, formatstr)
             else:
+                # remove and add again the keyframes so they are set on the
+                # current controller
+                keyframes = list(interpolator.keyframes)
+                for keyframe in keyframes:
+                    interpolator.removeKeyframe(keyframe)
+
                 interpolator.attachToElementProperty(gst_object_property,
                         gst_object)
                 interpolator.updateMediaStop(self.out_point)
 
-                # remove and add again the keyframes so they are set on the
-                # current controller
-                for keyframe in list(interpolator.keyframes):
-                    interpolator.removeKeyframe(keyframe)
+                for keyframe in keyframes:
                     interpolator.newKeyframe(keyframe)
 
             self.interpolators[gst_object_property.name] = \
@@ -501,7 +510,7 @@ class TrackObject(Signallable, Loggable):
     rate = property(_getRate)
 
     def _getPriority(self):
-        return self.gnl_object.props.priority
+        return self._public_priority
 
     def setPriority(self, priority):
         if self.timeline_object is not None:
@@ -510,11 +519,24 @@ class TrackObject(Signallable, Loggable):
             self.setObjectPriority(priority)
 
     def setObjectPriority(self, priority):
-        if self.gnl_object.props.priority != priority:
-            self.gnl_object.props.priority = priority
+        if priority != self._public_priority:
+            self._updatePriority(priority)
+
+    def _updatePriority(self, priority):
+        if self.stream_type is VideoStream:
+            true_priority = 2 + self._stagger + (3 * priority)
+        elif self.stream_type is AudioStream:
+            true_priority  = 2 + (2 * self._stagger) + (4 * priority)
+        if self.gnl_object.props.priority != true_priority:
+            self.gnl_object.props.priority = true_priority
 
     priority = property(_getPriority, setPriority)
 
+    def _getStagger(self):
+        return self._stagger
+
+    stagger = property(_getStagger)
+
     def trimStart(self, position, snap=False):
         if self.timeline_object is not None:
             self.timeline_object.trimStart(position, snap)
@@ -585,6 +607,7 @@ class TrackObject(Signallable, Loggable):
 
         bin = self.factory.makeBin(self.stream)
         self.gnl_object.add(bin)
+        self._rebuild_interpolators = True
         self._maybeBuildInterpolators()
 
     def releaseBin(self):
@@ -616,7 +639,15 @@ class TrackObject(Signallable, Loggable):
             i.updateMediaStop(stop)
 
     def _notifyPriorityCb(self, obj, pspec):
-        self.emit('priority-changed', obj.props.priority)
+        if self.stream_type is VideoStream:
+            true_priority = obj.props.priority
+            public_priority = (true_priority - 2 - self._stagger) // 3
+        elif self.stream_type is AudioStream:
+            true_priority = obj.props.priority
+            public_priority = (true_priority - 2 - (2 * self._stagger))// 4
+        if self._public_priority != public_priority:
+            self._public_priority = public_priority
+            self.emit('priority-changed', public_priority)
 
     def _connectToSignals(self, gnl_object):
         gnl_object.connect('notify::start', self._notifyStartCb)
@@ -641,6 +672,13 @@ class TrackObject(Signallable, Loggable):
     def _makeGnlObject(self):
         raise NotImplementedError()
 
+    def updatePosition(self, position):
+        if position != self._position:
+            self._position = position
+            self._stagger = position & 1
+            self._updatePriority(self._public_priority)
+            self.emit("stagger-changed", self._stagger)
+
 class SourceTrackObject(TrackObject):
 
     numobjs = 0
@@ -652,6 +690,181 @@ class SourceTrackObject(TrackObject):
         SourceTrackObject.numobjs += 1
         return source
 
+class Transition(Signallable):
+
+    __signals__ = {
+        "start-changed" : ["start"],
+        "duration-changed" : ["duration"],
+        "priority-changed" : ["priority"],
+    }
+
+    def __init__(self, a, b):
+        self.a = a
+        self.b = b
+        self.start = 0
+        self.duration = 0
+        self.priority = 0
+        self._makeGnlObject()
+        self._connectToTrackObjects(a, b)
+
+    def _makeGnlObject(self):
+        pass
+
+    def _connectToTrackObjects(self, a, b):
+        a.connect("start-changed", self._updateStartDuration)
+        a.connect("duration-changed", self._updateStartDuration)
+        a.connect("priority-changed", self._updatePriority)
+        b.connect("start-changed", self._updateStartDuration)
+        b.connect("priority-changed", self._updatePriority)
+        a.connect("stagger-changed", self._staggerChanged)
+        b.connect("stagger-changed", self._staggerChanged)
+        self._updateStartDuration()
+        self._updatePriority()
+
+    def _updateStartDuration(self, *unused):
+        start = self.b.start
+        end = self.a.start + self.a.duration
+        duration = max(0, end - start)
+
+        if start != self.start:
+            self._updateOperationStart(start)
+            self.start = start
+            self.emit("start-changed", start)
+
+        if duration != self.duration:
+            self._updateOperationDuration(duration)
+            self.duration = duration
+            self.emit("duration-changed", duration)
+
+        self._updateController()
+
+    def _updatePriority(self, *unused):
+        if self.a.priority == self.b.priority:
+            priority = self.a.priority
+            self._updateOperationPriority(priority)
+            self.priority = priority
+            self.emit("priority-changed", priority)
+
+    def _staggerChanged(self, *unused):
+        self._updateController()
+
+    def _updateController(self):
+        pass
+
+    def _updateOperationStart(self, start):
+        pass
+
+    def _updateOperationDuration(self, duration):
+        pass
+
+    def _updateOperationPriority(self, priority):
+        pass
+
+    def addThyselfToComposition(self, composition):
+        pass
+
+    def removeThyselfFromComposition(self, composition):
+        pass
+
+class VideoTransition(Transition):
+
+    caps = gst.Caps("video/x-raw-yuv,format=(fourcc)AYUV")
+
+    def _makeGnlObject(self):
+        trans = gst.element_factory_make("alpha")
+        self.controller = gst.Controller(trans, "alpha")
+        self.controller.set_interpolation_mode("alpha", gst.INTERPOLATE_LINEAR)
+
+        self.operation = gst.element_factory_make("gnloperation")
+        self.operation.add(trans)
+        self.operation.props.media_start = 0
+        self.operation.props.caps = self.caps
+
+    def addThyselfToComposition(self, composition):
+        composition.add(self.operation)
+
+    def removeThyselfFromComposition(self, composition):
+        composition.remove(self.operation)
+        self.operation.set_state(gst.STATE_NULL)
+
+    def _updateOperationStart(self, start):
+        self.operation.props.start = start
+
+    def _updateOperationDuration(self, duration):
+        self.operation.props.duration = duration
+        self.operation.props.media_duration = duration
+
+    def _updateOperationPriority(self, priority):
+        self.operation.props.priority = 1 + 3 * priority
+
+    def _updateController(self):
+        if self.a.stagger > self.b.stagger:
+            # source a is under source b (higher priority)
+            # we fade source B in
+            self.controller.unset_all("alpha")
+            self.controller.set("alpha", 0, 0.0)
+            self.controller.set("alpha", self.duration, 1.0)
+        elif self.a.stagger < self.b.stagger:
+            # source a is over source b (lower priority)
+            # we fade source a out
+            self.controller.unset_all("alpha")
+            self.controller.set("alpha", 0, 1.0)
+            self.controller.set("alpha", self.duration, 0.0)
+
+class AudioTransition(Transition):
+
+    def _makeGnlObject(self):
+        trans = gst.element_factory_make("volume")
+        self.a_controller = gst.Controller(trans, "volume")
+        self.a_controller.set_interpolation_mode("volume", gst.INTERPOLATE_LINEAR)
+
+        self.a_operation = gst.element_factory_make("gnloperation")
+        self.a_operation.add(trans)
+
+        trans = gst.element_factory_make("volume")
+        self.b_controller = gst.Controller(trans, "volume")
+        self.b_controller.set_interpolation_mode("volume", gst.INTERPOLATE_LINEAR)
+
+        self.b_operation = gst.element_factory_make("gnloperation")
+        self.b_operation.add(trans)
+
+        self.a_operation.props.media_start = 0
+
+        self.b_operation.props.media_start = 0
+
+
+    def addThyselfToComposition(self, composition):
+        composition.add(self.a_operation, self.b_operation)
+
+    def removeThyselfFromComposition(self, composition):
+        composition.remove(self.a_operation)
+        composition.remove(self.b_operation)
+        self.a_operation.set_state(gst.STATE_NULL)
+        self.b_operation.set_state(gst.STATE_NULL)
+
+    def _updateOperationStart(self, start):
+        self.a_operation.props.start = start
+        self.b_operation.props.start = start
+
+    def _updateOperationDuration(self, duration):
+        self.a_operation.props.duration = duration
+        self.a_operation.props.media_duration = duration
+        self.b_operation.props.duration = duration
+        self.b_operation.props.media_duration = duration
+
+    def _staggerChanged(self, *unused):
+        self._updateOperationPriority(self.priority)
+
+    def _updateOperationPriority(self, priority):
+        self.a_operation.props.priority = 1 + 2 * self.a.stagger + 4 * priority
+        self.b_operation.props.priority = 1 + 2 * self.b.stagger + 4 * priority
+
+    def _updateController(self):
+        self.a_controller.set("volume", 0, 1.0)
+        self.a_controller.set("volume", self.duration, 0.0)
+        self.b_controller.set("volume", 0, 0.0)
+        self.b_controller.set("volume", self.duration, 1.0)
+
 class Track(Signallable, Loggable):
     logCategory = "track"
 
@@ -660,15 +873,23 @@ class Track(Signallable, Loggable):
         'duration-changed': ['duration'],
         'track-object-added': ['track_object'],
         'track-object-removed': ['track_object'],
-        'max-priority-changed': ['track_object']
+        'max-priority-changed': ['track_object'],
+        'transition-added' : ['transition'],
+        'transition-removed' : ['transition'],
     }
 
     def __init__(self, stream):
         self.stream = stream
+        if type(self.stream) is VideoStream:
+            self.TransitionClass = VideoTransition
+        elif type(self.stream) is AudioStream:
+            self.TransitionClass = AudioTransition
         self.composition = gst.element_factory_make('gnlcomposition')
         self.composition.connect('notify::start', self._compositionStartChangedCb)
         self.composition.connect('notify::duration', self._compositionDurationChangedCb)
         self.track_objects = []
+        self.transitions = {}
+        self._update_transitions = True
         self._max_priority = 0
 
         self.mixer = self._getMixerForStream(stream)
@@ -684,8 +905,8 @@ class Track(Signallable, Loggable):
         else:
             return None
 
-        ret.priority = 2 ** 32 - 1
         ret.makeBin()
+        ret.gnl_object.props.priority = 2 ** 32 - 1
 
         return ret
 
@@ -759,7 +980,7 @@ class Track(Signallable, Loggable):
             return gnl
         elif isinstance(stream, VideoStream):
             gnl = gst.element_factory_make("gnloperation", "top-level-video-mixer")
-            m = SmartVideomixerBin()
+            m = SmartVideomixerBin(self)
             gnl.add(m)
             gnl.props.expandable = True
             gnl.props.priority = 0
@@ -802,12 +1023,6 @@ class Track(Signallable, Loggable):
 
     max_priority = property(_getMaxPriority)
 
-    _max_priority = 0
-
-    @property
-    def max_priority(self):
-        return self._max_priority
-
     def _trackObjectPriorityCb(self, trackobject, priority):
         op = self._max_priority
         self._max_priority = max((obj.priority for obj in self.track_objects))
@@ -844,6 +1059,8 @@ class Track(Signallable, Loggable):
         self._connectToTrackObject(track_object)
 
         self.emit('track-object-added', track_object)
+        if self._update_transitions:
+            self.updateTransitions()
 
     def removeTrackObject(self, track_object):
         if track_object.track is None:
@@ -860,12 +1077,15 @@ class Track(Signallable, Loggable):
 
         self.track_objects.remove(track_object)
         track_object.track = None
+
         self._disconnectTrackObjectSignals(track_object)
 
         self._updateMaxPriority()
         self.updateDefaultSources()
 
         self.emit('track-object-removed', track_object)
+        if self._update_transitions:
+            self.updateTransitions()
 
     def removeAllTrackObjects(self):
         for track_object in list(self.track_objects):
@@ -916,6 +1136,97 @@ class Track(Signallable, Loggable):
     def enableUpdates(self):
         self.composition.props.update = True
         self.updateDefaultSources()
+        self.updateTransitions()
 
     def disableUpdates(self):
         self.composition.props.update = False
+
+    def addTransition(self, transition):
+        a, b = transition.a, transition.b
+        if not ((a in self.track_objects) and
+                (b in self.track_objects)):
+            raise TrackError("One or both track objects not in track")
+        if (a, b) in self.transitions:
+            raise TrackError(
+                "A transition is already defined for these objects")
+        transition.addThyselfToComposition(self.composition)
+        self.transitions[a, b] = transition
+        self.emit("transition-added", transition)
+
+    def removeTransition(self, transition):
+        a, b = transition.a, transition.b
+        transition.removeThyselfFromComposition(self.composition)
+        del self.transitions[a, b]
+        self.emit("transition-removed", transition)
+
+    def getTrackObjectsGroupedByLayer(self):
+        layers = [[] for x in xrange(0, self.max_priority + 1)]
+        for track_object in self.track_objects:
+            layers[int(track_object.priority)].append(track_object)
+        return layers
+
+    def getValidTransitionSlots(self, objs):
+        prev = None
+        safe = 0
+        duration = 0
+        slots = []
+        valid = True
+
+        def pop():
+            if len(slots):
+                slots.pop(-1)
+        for obj in objs:
+            end = obj.start + obj.duration
+            if obj.start >= duration:
+                safe = obj.start
+                duration = end
+                prev = obj
+            elif end >= duration and obj.start >= safe:
+                slots.append((prev, obj))
+                safe = duration
+                duration = end
+                prev = obj
+            elif end >= duration and obj.start < safe:
+                pop()
+                valid = False
+                safe = duration
+                duration = end
+                prev = obj
+            elif end < duration and obj.start >= safe:
+                safe = end
+                valid = False
+            elif end < duration and obj.start < safe:
+                pop()
+                valid = False
+                safe = end
+        return slots, valid
+
+    valid_arrangement = True
+
+    def updateTransitions(self):
+        # create all new transitions
+        valid_slots = set()
+        all_valid = True
+        for layer in self.getTrackObjectsGroupedByLayer():
+            pos = 0
+            prev = None
+            slots, is_valid = self.getValidTransitionSlots(layer)
+            all_valid &= is_valid
+            for slot in slots:
+                a, b = slot
+                if a == prev:
+                    b.updatePosition(pos)
+                    pos += 1
+                else:
+                    a.updatePosition(pos)
+                    b.updatePosition(pos + 1)
+                    pos += 2
+                prev = b
+                valid_slots.add(slot)
+                if not slot in self.transitions:
+                    tr = self.TransitionClass(a, b)
+                    self.addTransition(tr)
+        current_slots = set(self.transitions.iterkeys())
+        for slot in current_slots - valid_slots:
+            self.removeTransition(self.transitions[slot])
+        self.valid_arrangement = all_valid
